import { Effect, Sink, Stream } from "effect";
import { RawAccount } from "cojson";
import { Account, AccountCtx, Group, SubscriptionScope, Ref, inspect, subscriptionsScopes, } from "../internal.js";
export function isCoValue(value) {
    return value && value._type !== undefined;
}
export function isCoValueClass(value) {
    return typeof value === "function" && value.fromRaw !== undefined;
}
/** @internal */
export class CoValueBase {
    get _owner() {
        const owner = this._raw.group instanceof RawAccount
            ? Account.fromRaw(this._raw.group)
            : Group.fromRaw(this._raw.group);
        const subScope = subscriptionsScopes.get(this);
        if (subScope) {
            subScope.onRefAccessedOrSet(owner.id);
            subscriptionsScopes.set(owner, subScope);
        }
        return owner;
    }
    /** @private */
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    constructor(..._args) { }
    /** @category Internals */
    static fromRaw(raw) {
        return new this({ fromRaw: raw });
    }
    /** @category Subscription & Loading */
    static loadEf(id) {
        return Effect.gen(this, function* (_) {
            const account = yield* _(AccountCtx);
            return yield* _(new Ref(id, account, this).loadEf());
        });
    }
    /** @category Subscription & Loading */
    static load(id, options) {
        return new Ref(id, options.as, this).load(options?.onProgress && { onProgress: options.onProgress });
    }
    /** @category Subscription & Loading */
    static subscribe(id, options, onUpdate) {
        void Effect.runPromise(Effect.provideService(this.subscribeEf(id).pipe(Stream.run(Sink.forEach((update) => Effect.sync(() => onUpdate(update))))), AccountCtx, options.as));
        return function unsubscribe() { };
    }
    /** @category Subscription & Loading */
    static subscribeEf(id) {
        return Stream.fromEffect(this.loadEf(id)).pipe(Stream.flatMap((value) => Stream.asyncScoped((emit) => Effect.gen(this, function* (_) {
            const subscription = new SubscriptionScope(value, this, (update) => {
                void emit.single(update);
            });
            yield* _(Effect.addFinalizer(() => Effect.sync(() => subscription.unsubscribeAll())));
        }))));
    }
    /** @category Subscription & Loading */
    subscribe(listener) {
        return this.constructor.subscribe(this.id, { as: this._loadedAs }, listener);
    }
    /** @category Subscription & Loading */
    subscribeEf() {
        return Stream.provideService(this.constructor.subscribeEf(this.id), AccountCtx, this._loadedAs);
    }
    toJSON() {
        return {
            id: this.id,
            type: this._type,
            error: "unknown CoValue class",
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    /** @category Type Helpers*/
    as(otherSchema) {
        const cast = otherSchema.fromRaw(this._raw);
        const subScope = subscriptionsScopes.get(this);
        if (subScope) {
            subscriptionsScopes.set(cast, subScope);
        }
        return cast;
    }
}
//# sourceMappingURL=interfaces.js.map